import os
import boto
from boto.s3.key import Key

def push_s3(azn_key, azn_secret, s3bucket, localdir, s3dir):
    """Recurses through localdir, uploading each file it finds into the corresponding
    path in s3dir"""
    conn = boto.connect_s3(azn_key, azn_secret)
    bucket = conn.get_bucket(s3bucket)
    tree = os.walk(localdir)
    for folder in tree:
        folderpath = folder[0]
        subfiles = folder[2]
        for file in subfiles:
            localfile = os.path.join(folderpath, file)
            trimmed_dir = folderpath[len(localdir):]
            #os.path.normpath cleans out any double slashes //
            #warning: if run from windows the slashes will get turned into backslashes
            s3file = os.path.normpath("/".join((s3dir, trimmed_dir, file)))
            print "pushing %s to %s" % (localfile, s3file)
            k = Key(bucket)
            k.key = s3file
            k.set_contents_from_filename(localfile)
            k.set_acl('public-read')



#!/usr/bin/env python
import subprocess
import sys
import time
import os
import pkg_resources
import shutil
import random
from signal import SIGTERM, SIGINT

import cherrypy
import yaml

import silk.lib

def run_til_you_die(args, kill_signal, cwd=os.getcwd(), env={}):
    print 'Running %s in dir %s with env %s' % (' '.join(args), cwd, str(env))
    env.update(os.environ)
    proc = subprocess.Popen(args, cwd=cwd, env=env)
    try:
        proc.wait()
    except KeyboardInterrupt as e:
        print "KeyboardInterrupt"
        proc.send_signal(kill_signal)
    except Exception as e:
        print e
        proc.send_signal(kill_signal)

def run_fab(args):
    args[0] = 'fab'
    run_til_you_die(args, SIGTERM)

def run_gunicorn(config, root):
    cmd = silk.lib.get_gunicorn_cmd(config)
    print 'config', config
    run_til_you_die(cmd.split(), SIGINT, cwd=root, env=config['env'])
    #give the child processes some time to end before returning the shell
    #there must be a more elegant way to do that, but I don't know it
    time.sleep(1)

def run_cherrypy(config, root):
    """Mount each of the static dirs in site.yaml as its own app
    in the cherrypy tree.  Then mount the wsgi app as the root, and run."""
    if config['static_dirs']:
        for static_dir in config['static_dirs']:
            #mount each of our static dirs as its own app in the cherrypy tree
            url_path = static_dir['url_path'].rstrip('/')
            sys_path = os.path.join(root, static_dir['system_path'])
            cherry_conf = {'/': {'tools.staticdir.on': True,
                                 'tools.staticdir.dir': sys_path,}}
            cherrypy.tree.mount(None, script_name=url_path, config=cherry_conf)
    #mount the wsgi app
    sys.path.append(root)
    os.chdir(root)
    wsgi_app = cherrypy.lib.attributes(config['wsgi_app'].replace(':', '.'))
    cherrypy.tree.graft(wsgi_app, '')
    #update the env vars with the ones from the deployment role
    os.environ.update(config['env'])
    #run cherrypy
    cherrypy.config.update({
        'server.socket_host': '127.0.0.1',
        'server.socket_port': 8000
    })
    cherrypy.engine.start()
    cherrypy.engine.block()

def install_skel(sitename):
    """Copies the contents of site_templates into the named directory (within cwd)"""
    #get the dir from pkg_resources
    src = pkg_resources.resource_filename('silk', 'site_templates')
    try:
        shutil.copytree(src, os.path.join(os.getcwd(), sitename))
    except OSError, e:
        print e
    
def freeze_2_yaml():
    """Read lines of text from stdin and print a python_packages yaml list"""
    lines = sys.stdin.read().split('\n')#split only on newlines, not spaces
    lines = [line for line in lines if line]#remove empty lines
    print yaml.safe_dump({'python_packages':lines}, default_flow_style=False)

def yaml_2_freeze():
    """Read lines of deps.yaml from stdin and print requirements.txt contents"""
    txt = sys.stdin.read()
    deps = yaml.safe_load(txt)
    print '\n'.join(deps['python_packages'])

def local_python_deps():
    """Write a requirements.txt from deps.yaml file for pip, then run pip on it."""
    depfile = os.path.join(silk.lib.get_site_root(os.getcwd()), 'deps.yaml')
    txt = open(depfile).read()
    deps = yaml.safe_load(txt)
    print deps
    reqs = '\n'.join(deps['python_packages'])
    print reqs
    tmpfile = ''.join([random.choice('abcdefg') for x in xrange(5)]) + '-reqs.txt'
    #OH NOE, LINUX-SPECIFIC
    tmpfile = os.path.join('/tmp', tmpfile)
    f = open(tmpfile, 'w')
    f.write(reqs)
    f.close()
    cmd = 'pip install -r %s' % tmpfile
    run_til_you_die(cmd.split(), SIGTERM)
    os.remove(tmpfile)

cmd_map = {
    'grun': run_gunicorn,
    'run': run_cherrypy,
    'skel': install_skel,
    'deps': local_python_deps
}

config_required = (
    run_gunicorn,
    run_cherrypy,
)

def cmd_dispatcher():
    """wraps 'fab', handles 'silk run'"""
    args = sys.argv
    try:
        cmd = args[1]
    except IndexError:
        run_fab(*args)

    if cmd in cmd_map:
        cmd_func = cmd_map[cmd]
        if cmd_func in config_required:
            role = silk.lib.get_role()
            root = silk.lib.get_site_root(os.getcwd())
            config = silk.lib.get_config(root, role)
            cmd_func(config, root)
        else:
            cmd_func(*args[2:])
    else:
        run_fab(args)



from silk.fabfile import push as real_push
from silk.utils import run_til_you_die as _run_til_you_die
from signal import SIGINT
import silk
import os
#SIGH

#time to rewrite code that I wrote last night but is trapped on my laptop because I didn't do an hg addremove
#SIGH
#I think the only new file was django.py.
#which includes collectstatic, and an overwritten push

def _get_proj_dir(root):
    """Ugly magic function to loop through subdirectories and return the first
    one that looks like a Django project (has a settings.py)"""
    #get list of current folder contents
    paths = [os.path.join(root, folder) for folder in os.listdir(root) if os.path.isdir(os.path.join(os.path.join(root, folder)))]
    for path in paths:
        if os.path.isfile(os.path.join(path, 'settings.py')):
            return path

def collectstatic():
    """Runs ./manage.py collectstatic, setting current Silk role 
    as env var so it can be picked up in local_settings"""
    args = ['./manage.py', 'collectstatic', '--settings=local_settings']
    proj_dir = _get_proj_dir(silk.lib.get_site_root(os.getcwd()))
    env_vars = {'SILK_ROLE': silk.lib.get_role()}
    _run_til_you_die(args, SIGINT, proj_dir, env=env_vars)

def push():
    collectstatic()
    real_push()



import yaml
import copy
import os
import sys
import pkg_resources

def get_gunicorn_cmd(site_config, bin_dir=''):
    gconfig = copy.copy(site_config['gunicorn'])
    debug = gconfig.pop('debug', None)
    options = ' '.join(['--%s %s' % (x, y) for x, y in gconfig.iteritems()])
    if debug:
        options += ' --debug'
    gconfig['options'] = options
    gconfig['bin_dir'] = bin_dir
    gconfig.update(site_config)
    cmd = 'gunicorn %(options)s %(wsgi_app)s' % gconfig
    if bin_dir:
        cmd = '%s/%s' % (bin_dir, cmd)
    return cmd

def get_site_root(start_dir):
    testfile = os.path.join(start_dir, 'site.yaml')
    if os.path.isfile(testfile):
        return start_dir
    else:
        parent_dir = os.path.split(start_dir)[0]
        if parent_dir != start_dir:
            return get_site_root(parent_dir)
        else:
            return None

def get_role_list(local_root):
    """Return a list of the role names defined by yaml roles/*.yaml"""
    return [file[:-5] for file in os.listdir(os.path.join(local_root, 'roles')) if file.endswith('.yaml')]

def get_role_config(role):
    role_file = '%s/roles/%s.yaml' % (get_site_root(os.getcwd()), role)
    config =  yaml.safe_load(open(role_file, 'r').read())
    return config
    #TODO: support pulling role info from a web page

def get_site_config(site_root):
    """Parses and returns site.yaml"""
    site_config_file = os.path.join(site_root, 'site.yaml')
    config = yaml.safe_load(open(site_config_file, 'r').read())
    return config

def get_blame(site_root):
    """Parses and returns blame.yaml in deployed site"""
    blame_file = os.path.join(site_root, 'blame.yaml')
    blame = yaml.safe_load(open(blame_file, 'r').read())
    return blame

def get_config(site_root, role=None):
    """Returns merged site and role config.
    Falls back to blame.yaml if no role given.
    Falls back to just site.yaml if no role given and no blame file found"""
    if role is None:
        try:
            return get_blame(site_root)[1]['config']
        except IOError:
            return get_site_config(site_root)
    else:
        config = get_site_config(site_root)
        config.update(get_role_config(role))
        return config

def get_template_path(template, site_root=None):
    """
    Returns path of template from site cfg_templates dir, if found there, else
    returns template path from silk's cfg_templates dir.
    """
    if site_root:
        localpath=os.path.join(site_root, 'cfg_templates', template)
        if os.path.isfile(localpath):
            return localpath
    pkgpath=pkg_resources.resource_filename('silk', 'cfg_templates/%s' % template)
    if os.path.isfile(pkgpath):
        return pkgpath
    else:
        raise Exception("Template not found: %s" % template)

def get_rendered_template(template_name, context):
    """
    Returns text of named template, with keyword substitutions pulled from
    'context'
    """
    template_path = get_template_path(template_name)
    txt = open(template_path, 'r').read()
    return txt % context

def get_role():
    try:
        #if '-R rolename' found in sys.argv, use that
        return sys.argv[sys.argv.index('-R')+1]
    except:
        #role not found in sys.argv, try env var
        #return None if no role there either
        return os.environ.get('SILK_ROLE', None)



import sys
import os
import datetime
import pkg_resources
import random
import posixpath

import yaml
from fabric.api import *
from fabric.contrib.files import exists, upload_template

import silk.lib

SRV_ROOT = '/srv'
ROLLBACK_CAP = 3

def _get_silk_deps():
    silk_deps_file = pkg_resources.resource_filename('silk', 'deps.yaml')
    return yaml.safe_load(open(silk_deps_file).read())

def _get_site_deps(local_root):
    site_deps_file = os.path.join(local_root, 'deps.yaml')
    return yaml.safe_load(open(site_deps_file).read())

def _set_vars():
    """
    Loads deployment settings into Fabric's global 'env' dict
    """
    env.local_root = silk.lib.get_site_root(os.getcwd())
    env.config = silk.lib.get_site_config(env.local_root)
    if len(env.roles) == 0:
        sys.exit("ERROR: you must define a role with -R <rolename>")
    elif len(env.roles) > 1:
        sys.exit("ERROR: Silk only permits passing in one role at a time")
    else:
        env.config.update(silk.lib.get_role_config(env.roles[0]))
    env.site = env.config['site']
    env.remote_root = '/'.join([SRV_ROOT, env.site])
    env.envdir = '/'.join([env.remote_root, 'env'])
    env.workdir = '/'.join(['/tmp', env.site])
    env.rollbackdir = '/'.join([SRV_ROOT, 'rollbacks'])
    env.deploytime = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
    #default to using a unix socket for nginx->gunicorn
    if 'bind' not in env.config['gunicorn']:
        env.config['gunicorn']['bind'] = 'unix:/tmp/%s.sock' % env.config['site']
    env.config['bind'] = env.config['gunicorn']['bind']
    env.config['silk_deps'] = _get_silk_deps()
    env.config['site_deps'] = _get_site_deps(env.local_root)

_set_vars()

# UGLY MAGIC
# Here we're (ab)using Fabric's built in 'role' feature to work with the way 
# we're loading context-specific config.  Could possibly use a refactor to 
# avoid Fabric roles altogether.
def _get_hosts():
    """Return list of hosts to push to"""
    return env.config['push_hosts']

for role in silk.lib.get_role_list(env.local_root):
    env.roledefs[role] = _get_hosts
# END UGLY MAGIC

def _put_dir(local_dir, remote_dir):
    """
    Copies a local directory to a remote one, using tar and put. Silently
    overwrites remote directory if it exists, creates it if it does not
    exist.
    """
    local_tgz = "/tmp/fabtemp.tgz"
    remote_tgz = os.path.basename(local_dir) + ".tgz"
    local('tar -C "{0}" -czf "{1}" .'.format(local_dir, local_tgz))
    put(local_tgz, remote_tgz)
    local('rm -f "{0}"'.format(local_tgz))
    run('rm -Rf "{0}"; mkdir -p "{0}"; tar -C "{0}" -xzf "{1}" && rm -f "{1}"'\
        .format(remote_dir, remote_tgz))

def _get_blame():
    """
    Returns a yaml file that contains the site config, plus some deployment info
    """
    blame = [
        {'deployed_by': env.user,
        'deployed_from': os.uname()[1],
        'deployed_at': datetime.datetime.now(),
         'deployed_role': env.roles[0]},
        {'config':env.config}
    ]
    return yaml.safe_dump(blame, default_flow_style=False)

def _write_blame():
    """
    Writes blame file on remote host.
    """
    #randomize the string to reduce chance of collision
    blamefile = '/tmp/%s-blame.yaml' % ''.join([random.choice('abcdefghijklmnopqrstuvwxyz') for x in xrange(5)])
    f = open(blamefile, 'w')
    f.write(_get_blame())
    f.close()
    put(blamefile, '/'.join([env.workdir, 'blame.yaml']))
    os.remove(blamefile)

def reload():
    """
    Reloads supervisord and nginx configs.
    """
    print "RELOADING CONFIGS"
    sudo('supervisorctl reload')
    sudo('/etc/init.d/nginx reload')

def restart():
    """
    Restarts nginx and supervisord.  Normally not needed (reload() is enough)
    """
    print "RESTARTING SERVICES"
    sudo('/etc/init.d/supervisor stop; /etc/init.d/supervisor start')
    sudo('/etc/init.d/nginx restart')

def archive():
    """
    Creates rollback archive of already-deployed site.  Rotates old rollback files.
    """
    print "CREATING ROLLBACK"
    if not exists(env.rollbackdir, use_sudo=True):
        sudo('mkdir -p %s' % env.rollbackdir)

    template_vars = {
        'rollback_cap': ROLLBACK_CAP,
        'srv_root': SRV_ROOT,
    }

    template_vars.update(env)

    oldest_rollback = '%(rollbackdir)s/%(site)s-rollback_%(rollback_cap)s.tar.bz2' % template_vars
    #first delete the oldest rollback if present
    if exists(oldest_rollback):
        sudo('rm %s' % oldest_rollback) 

    #then increment the numbers on the existing rollbacks
    for i in xrange(ROLLBACK_CAP - 1, 0, -1):
        rollback_file = '%s/%s-rollback_%s.tar.bz2' % (env.rollbackdir, env.site, i)
        if exists(rollback_file):
            newname = '%s/%s-rollback_%s.tar.bz2' % (env.rollbackdir, env.site, i + 1)
            sudo('mv %s %s' % (rollback_file, newname))

    #now archive env.remote_root if it exists
    if exists(env.remote_root):
        sudo('tar -cjf %(rollbackdir)s/%(site)s-rollback_1.tar.bz2 --exclude "*.log" -C %(srv_root)s %(site)s' % template_vars)

def rollback():
    """
    Untars most recent rollback archive and sets it running.
    """
    print "ROLLING BACK"
    rollback_file = '%s/%s-rollback_1.tar.bz2' % (env.rollbackdir, env.site)
    if exists(rollback_file): 
        #unzip in a tmp dir
        tmpdir = os.path.join('/tmp', 'rollback-%s' % env.site)
        if exists(tmpdir, use_sudo=True):
            sudo('rm %s -rf' % tmpdir)
        sudo('mkdir %s' % tmpdir)
        sudo('tar -xjf %s -C %s' % (rollback_file, tmpdir))

        #move current code into oldddir
        olddir = os.path.join('/tmp', 'old-%s' % env.site)
        if exists(env.remote_root, use_sudo=True):
            sudo('mv %s %s' % (env.remote_root, olddir))

        #move new code into srvdir
        sudo('mv %s/%s %s' % (tmpdir, env.site, env.remote_root))

        #remove olddir
        if exists(olddir, use_sudo=True):
            sudo('rm %s -rf' % olddir)

        #clean out the rollback file we just unzipped
        sudo('rm -rf %s' % rollback_file)

        #decrement the other rollback files
        for i in xrange(2, ROLLBACK_CAP + 1, 1):
            oldname = '%s/%s-rollback_%s.tar.bz2' % (env.rollbackdir, env.site, i)
            newname = '%s/%s-rollback_%s.tar.bz2' % (env.rollbackdir, env.site, i - 1)
            if exists(oldname):
                sudo('mv %s %s' % (oldname, newname))
        reload()
    else:
        sys.exit('Error: %s not found' % rollback_file)


def install_apt_deps():
    """
    Installs system packages and build dependencies with apt.
    """
    for deps_dict in (env.config['silk_deps'], env.config['site_deps']):
        print deps_dict
        if deps_dict['apt_build_deps']:
            sudo('apt-get build-dep %s -y' % ' '.join(deps_dict['apt_build_deps']))
        if deps_dict['apt_packages']:
            sudo('apt-get install %s -y' % ' '.join(deps_dict['apt_packages']))

#TODO: rebuild virtualenv if it exists but the python version is wrong
def create_virtualenv():
    """
    Creates a virtualenv for the site.  Automatically builds egenix-mx-tools in it, since
    pip doesn't seem able to install that.
    """
    print "CREATING VIRTUALENV"
    if not exists(env.remote_root, use_sudo=True):
        sudo('mkdir -p %s' % env.remote_root)
    sudo('virtualenv --no-site-packages --python=%s %s' % (env.config['runtime'], env.envdir))
    build_mx_tools()

def install_python_deps():
    """
    Runs pip install inside the remote virtualenv.  First for silk's dependencies and
    then for the site's.
    """
    for dep_dict in (env.config['silk_deps'], env.config['site_deps']):
        if dep_dict['python_packages']:
            sudo('%s/bin/pip install %s' % (
                env.envdir, 
                ' '.join(dep_dict['python_packages']),
            ))

def install_deps():
    """
    Wraps the apt deps, virtualenv creation, and python deps functions to ensure
    that things are done in the right order.
    """
    print "INSTALLING DEPENDENCIES"
    install_apt_deps()
    if not exists(env.envdir, use_sudo=True):
        create_virtualenv()
    install_python_deps()

def build_mx_tools():
    """
    Builds and install egenix-mx-tools into virtualenv
    """
    #egenix-mx-tools includes the mxdatetime module, which is
    #a psycopg2 dependency.  Unfortunately it's not packaged in 
    #a way that pip can install.  So we build it here instead
    print "INSTALLING MX TOOLS"
    build_dir = "/tmp/egenix_build"
    if not exists(build_dir, use_sudo=True):
        sudo('mkdir -p %s' % build_dir)
    with cd(build_dir):
        #download the tools
        sudo('wget http://downloads.egenix.com/python/egenix-mx-base-3.1.3.tar.gz')
        #unpack
        sudo('tar xvf egenix-mx-base-3.1.3.tar.gz')
    with cd(os.path.join(build_dir, 'egenix-mx-base-3.1.3')):
        #install into the virtualenv
        sudo('%s setup.py install' % os.path.join(env.envdir, 'bin', 'python'))
    sudo('rm -rf %s' % build_dir)

def push_code():    
    """
    Pushes site to remote host
    """
    print "PUSHING CODE TO HOST"
    if exists(env.workdir):
        sudo('rm %s -rf' % env.workdir)
    _put_dir(env.local_root, env.workdir)

def _upload_config_template(template, dest, context):
    #first try to load the template from the local cfg_templates dir.
    #if it's not there then try loading from pkg_resources
    path = silk.lib.get_template_path(template)
    upload_template(path, dest, context=context, use_sudo=True)

def _get_nginx_static_snippet(url_path, local_path):
    return """
    location %(url_path)s {
        alias %(local_path)s;
    }
    """ % locals()

def _write_templates(template_list, template_vars):
    for pair in template_list:
        src, dest = pair
        _upload_config_template(
            src,
            dest,
            context = template_vars
        )

def _ensure_dir(remote_path):
    if not exists(remote_path, use_sudo=True):
        sudo('mkdir %s' % remote_path)

def _format_supervisord_env(env_dict):
    """Takes a dictionary and returns a string in form
    'key=val,key2=val2'"""
    try:
      return ','.join(['%s="%s"' % (key, env_dict[key]) for key in env_dict.keys()])
    except AttributeError:
      #env_dict isn't a dictionary, so they must not have included any env vars for us.
      #return empty string
      return ''

def write_config():
    """
    Creates and upload config files for nginx, supervisord, and blame.yaml
    """
    print "WRITING CONFIG"
    nginx_static = ''
    static_dirs = env.config.get('static_dirs', None)
    if static_dirs:
      for item in static_dirs:
          nginx_static += _get_nginx_static_snippet(
              item['url_path'],
              #system_path may be a full path, or relative to remote_root
              posixpath.join(env.remote_root, item['system_path'])
          )
    template_vars = {
        'cmd': silk.lib.get_gunicorn_cmd(env.config, bin_dir='%s/bin' % (env.envdir)),
        'nginx_static': nginx_static,
        'nginx_hosts': ' '.join(env.config['listen_hosts']),
        'process_env': _format_supervisord_env(env.config['env']),
        'srv_root': SRV_ROOT,
    }
    template_vars.update(env)
    template_vars.update(env.config)
    config_dir = '/'.join([env.workdir, 'conf'])
    #make sure the conf and logs dirs are created
    _ensure_dir(config_dir)
    _ensure_dir('/'.join([env.workdir, 'logs']))

    template_list = (
        ('supervisord.conf','/'.join([config_dir, 'supervisord.conf'])),
        ('nginx.conf','/'.join([config_dir, 'nginx.conf'])),
    )
    _write_templates(template_list, template_vars)
    _write_blame()

def switch():
    """
    Does a little dance to move the old project dir out of the way and put
    the new one in its place.
    """
    print "MOVING NEW CODE INTO PLACE"
    #copy the virtualenv into env.workdir
    sudo('cp %s %s/ -r' % (env.envdir, env.workdir))
    #move old env.remote_root
    olddir = '/'.join(['/tmp', 'old-%s' % env.site])
    sudo('mv %s %s' % (env.remote_root, olddir))
    #move code into place
    sudo('mv %s %s' % (env.workdir, env.remote_root))

def cleanup():
    """
    Removes the old project dir.  (But you still have a rollback!)
    """
    print "CLEANING UP"
    #do this last to minimize time between taking down old site and setting up new one
    #since "mv" is faster than "cp -r" or "rm -rf"
    olddir = '/'.join(['/tmp', 'old-%s' % env.site])
    sudo('rm %s -rf' % olddir)

def server_setup():
    """
    Installs nginx and supervisord on remote host.
    Sets up nginx and supervisord global config files.
    """
    install_apt_deps()
    template_list = (
        ('supervisord_root.conf', '/etc/supervisor/supervisord.conf'),
        ('nginx_root.conf','/etc/nginx/nginx.conf'),
    )
    _write_templates(template_list, env)
    restart()

def push():
    """
    The main function.  Assuming you have nginx and supervisord installed
    and configured, this function will put your site on the remote host and
    get it running.
    """
    archive()
    install_deps()
    push_code()
    write_config()
    switch()
    reload()
    cleanup()

def dump_config():
    """
    Just loads up the config and prints it out. For testing.
    """
    print env.config



#!/usr/bin/python
from setuptools import setup, find_packages

setup(
    name='silk-deployment',
    version='0.1.1',
    author='Brent Tubbs',
    author_email='brent.tubbs@gmail.com',
	packages=find_packages(),
    package_dir={'silk': 'silk'},
    include_package_data=True,
    entry_points = {
        'console_scripts': [
            'silk = silk.utils:cmd_dispatcher',
            'freeze2yaml = silk.utils:freeze_2_yaml',
            'yaml2freeze = silk.utils:yaml_2_freeze',
        ],
    },
	install_requires = [
        'cherrypy',#for 'silk run' (local devserver)
        'Fabric >= 0.9.2',
        'PyYAML',
	],
    url='http://bits.btubbs.com/silk-deployment',
    license='LICENSE.txt',
    description='A Fabric-based tool for deploying WSGI apps on an Ubuntu/Nginx/Supervisord/Gunicorn stack.',
    long_description=open('README.rst').read(),
)

